@@ -2,12 +2,12 @@
 
 #  ************************** Copyrights and license ***************************
 #
-# This file is part of gcovr 8.2, a parsing and reporting tool for gcov.
-# https://gcovr.com/en/8.2
+# This file is part of gcovr 8.4, a parsing and reporting tool for gcov.
+# https://gcovr.com/en/8.4
 #
 # _____________________________________________________________________________
 #
-# Copyright (c) 2013-2024 the gcovr authors
+# Copyright (c) 2013-2025 the gcovr authors
 # Copyright (c) 2013 Sandia Corporation.
 # Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 # the U.S. Government retains certain rights in this software.
@@ -20,34 +20,36 @@
 # cspell:ignore getpreferredencoding getfixture caplog
 
 from __future__ import annotations
-from argparse import ArgumentParser, ArgumentTypeError, SUPPRESS
+from argparse import _ArgumentGroup, ArgumentParser, ArgumentTypeError, SUPPRESS
 from inspect import isclass
 from locale import getpreferredencoding
 import logging
-from typing import Iterable, Any, List, Optional, Callable, TextIO, Dict
+from typing import Iterable, Any, Optional, Callable, TextIO
 from dataclasses import dataclass
 import datetime
 import os
 import re
 
 from . import formats
+from .timestamps import parse_timestamp
 from .options import (
+    FilterOption,
     GcovrConfigOption,
     GcovrConfigOptionAction,
     GcovrDeprecatedConfigOptionAction,
+    NonEmptyFilterOption,
     Options,
     OutputOrDefault,
     check_input_file,
     check_percentage,
     relative_path,
 )
-from .utils import FilterOption
 
 LOGGER = logging.getLogger("gcovr")
 
 
 def timestamp(value: str) -> datetime.datetime:
-    from .timestamps import parse_timestamp  # lazy import
+    """Get the current timestamp from a given string."""
 
     try:
         return parse_timestamp(value)
@@ -56,8 +58,7 @@ def timestamp(value: str) -> datetime.datetime:
 
 
 def source_date_epoch() -> Optional[datetime.datetime]:
-    """
-    Load time from SOURCE_DATE_EPOCH, if it exists.
+    """Load time from SOURCE_DATE_EPOCH, if it exists.
     See: <https://reproducible-builds.org/docs/source-date-epoch/>
 
     Examples:
@@ -90,7 +91,7 @@ def source_date_epoch() -> Optional[datetime.datetime]:
     if ts:
         try:
             return datetime.datetime.fromtimestamp(int(ts), datetime.timezone.utc)
-        except Exception:
+        except ValueError:
             LOGGER.warning(
                 "Ignoring invalid environment variable SOURCE_DATE_EPOCH=%r",
                 ts,
@@ -99,7 +100,9 @@ def source_date_epoch() -> Optional[datetime.datetime]:
     return None
 
 
-def argument_parser_setup(parser: ArgumentParser, default_group):
+def argument_parser_setup(
+    parser: ArgumentParser, default_group: _ArgumentGroup
+) -> None:
     r"""Add all options and groups to the given argparse parser."""
 
     # setup option groups
@@ -115,13 +118,15 @@ def argument_parser_setup(parser: ArgumentParser, default_group):
     for opt in GCOVR_CONFIG_OPTIONS:
         group = default_group if opt.group is None else groups[opt.group]
 
-        kwargs: Dict[str, Any] = {
-            "action": opt.action,
-            "const": opt.const,
-            "default": SUPPRESS,  # default will be assigned manually
-            "help": opt.help,
-            "metavar": opt.metavar,
-        }
+        kwargs = dict[str, Any](
+            {
+                "action": opt.action,
+                "const": opt.const,
+                "default": SUPPRESS,  # default will be assigned manually
+                "help": opt.help,
+                "metavar": opt.metavar,
+            }
+        )
 
         # To avoid store_const problems, optionally set choices, nargs, type:
         if opt.choices is not None:
@@ -153,30 +158,30 @@ def argument_parser_setup(parser: ArgumentParser, default_group):
 
 def parse_config_into_dict(
     config_entry_source: Iterable[ConfigEntry],
-    all_options: Iterable[GcovrConfigOption] = None,
-) -> Dict[str, Any]:
-    cfg_dict: Dict[str, Any] = {}
+    all_options: Optional[Iterable[GcovrConfigOption]] = None,
+) -> dict[str, Any]:
+    """Parse a config file and save the configuration in a dictionary."""
+    cfg_dict = dict[str, Any]()
 
     if all_options is None:
         all_options = GCOVR_CONFIG_OPTIONS
 
-    options_lookup = {}
-    for option in all_options:
-        if option.config_keys is not None:
-            for config_key in option.config_keys:
-                options_lookup[config_key] = option
-
     for cfg_entry in config_entry_source:
-        try:
-            option: GcovrConfigOption = options_lookup[cfg_entry.key]
-        except KeyError:
+        for option in all_options:
+            if option.config_keys is not None:
+                if cfg_entry.key in option.config_keys:
+                    value = _get_value_from_config_entry(cfg_entry, option)
+                    _assign_value_to_dict(
+                        cfg_dict,
+                        value,
+                        option,
+                        cfg_entry_key=cfg_entry.key,
+                        is_single_value=True,
+                    )
+                    break
+        else:
             raise cfg_entry.error("unknown config option") from None
 
-        value = _get_value_from_config_entry(cfg_entry, option)
-        _assign_value_to_dict(
-            cfg_dict, value, option, cfg_entry_key=cfg_entry.key, is_single_value=True
-        )
-
     return cfg_dict
 
 
@@ -184,7 +189,7 @@ def _get_value_from_config_entry(
     cfg_entry: ConfigEntry,
     option: GcovrConfigOption,
 ) -> Any:
-    def get_boolean(silent_error: bool = False):
+    def get_boolean(silent_error: bool = False) -> Optional[bool]:
         try:
             return cfg_entry.value_as_bool
         except ValueError:
@@ -224,9 +229,9 @@ def _get_value_from_config_entry(
         try:
             value = converter(cfg_entry.value)
         except (ValueError, ArgumentTypeError) as err:
-            raise cfg_entry.error(str(err))
+            raise cfg_entry.error(str(err)) from None
 
-    elif option.name == "json_add_tracefile":  # Special case for patterns
+    elif option.name == "json_tracefile":  # Special case for patterns
         if cfg_entry.filename is None:
             raise AssertionError(
                 "Conversion function must derive base directory from filename"
@@ -276,11 +281,11 @@ def _get_converter_function(
 
 
 def _assign_value_to_dict(
-    namespace: Dict[str, Any],
+    namespace: dict[str, Any],
     value: Any,
     option: GcovrConfigOption,
     is_single_value: bool,
-    cfg_entry_key: str = None,
+    cfg_entry_key: Optional[str] = None,
 ) -> None:
     if option.action == "append" or option.nargs == "*":
         append_target = namespace.setdefault(option.name, [])
@@ -294,9 +299,13 @@ def _assign_value_to_dict(
         namespace[option.name] = value
         return
 
-    if issubclass(option.action, GcovrConfigOptionAction):
-        option.action(option.flags, option.name)(
-            None, namespace, value, config=cfg_entry_key
+    if not isinstance(option.action, str) and issubclass(
+        option.action, GcovrConfigOptionAction
+    ):
+        option.action(option.flags, option.name).store_config_key(
+            namespace,
+            value,
+            config=cfg_entry_key,
         )
         return
 
@@ -304,19 +313,19 @@ def _assign_value_to_dict(
 
 
 def merge_options_and_set_defaults(
-    partial_namespaces: List[Dict[str, Any]],
-    all_options: Optional[List[GcovrConfigOption]] = None,
+    partial_namespaces: list[dict[str, Any]],
+    all_options: Optional[list[GcovrConfigOption]] = None,
 ) -> Options:
+    """Merge all options into the namespace and set the default values for unused options."""
     if not partial_namespaces:
         raise AssertionError("At least one namespace required")
 
     if all_options is None:
         all_options = GCOVR_CONFIG_OPTIONS
 
-    target: Dict[str, Any] = {}
+    target = dict[str, Any]()
     for namespace in partial_namespaces:
         for option in all_options:
-
             if option.name not in namespace:
                 continue
 
@@ -332,12 +341,16 @@ def merge_options_and_set_defaults(
 
 
 class UseSortUncoveredNumberAction(GcovrDeprecatedConfigOptionAction):
+    """Argparse action to map old option --sort-uncovered to new option --sort=uncovered-number."""
+
     option = "--sort"
     config = "sort"
     value = "uncovered-number"
 
 
 class UseSortUncoveredPercentAction(GcovrDeprecatedConfigOptionAction):
+    """Argparse action to map old option --sort-percentage to new option --sort=uncovered-percent."""
+
     option = "--sort"
     config = "sort"
     value = "uncovered-percent"
@@ -445,6 +458,91 @@ GCOVR_CONFIG_OPTIONS = [
         ),
         action="store_false",
     ),
+    GcovrConfigOption(
+        "medium_threshold",
+        ["--medium-threshold", "--html-medium-threshold"],
+        group="output_options",
+        type=check_percentage,
+        metavar="MEDIUM",
+        help=(
+            "If the coverage is below MEDIUM, the value is marked "
+            "as low coverage in the report. "
+            "MEDIUM has to be lower than or equal to value of --high-threshold "
+            "and greater than 0. "
+            "If MEDIUM is equal to value of --high-threshold the report has "
+            "only high and low coverage. Default is {default!s}."
+        ),
+        default=75.0,
+    ),
+    GcovrConfigOption(
+        "high_threshold",
+        ["--high-threshold", "--html-high-threshold"],
+        group="output_options",
+        type=check_percentage,
+        metavar="HIGH",
+        help=(
+            "If the coverage is below HIGH, the value is marked "
+            "as medium coverage in the report. "
+            "HIGH has to be greater than or equal to value of --medium-threshold. "
+            "If HIGH is equal to value of --medium-threshold the report has "
+            "only high and low coverage. Default is {default!s}."
+        ),
+        default=90.0,
+    ),
+    GcovrConfigOption(
+        "medium_threshold_branch",
+        ["--medium-threshold-branch", "--html-medium-threshold-branch"],
+        group="output_options",
+        metavar="MEDIUM_BRANCH",
+        type=check_percentage,
+        help="If the coverage is below MEDIUM_BRANCH, the value is marked "
+        "as low coverage in the report. "
+        "MEDIUM_BRANCH has to be lower than or equal to value of --high-threshold-branch "
+        "and greater than 0. "
+        "If MEDIUM_BRANCH is equal to value of --medium-threshold-branch the report has "
+        "only high and low coverage. Default is taken from --medium-threshold.",
+        default=None,
+    ),
+    GcovrConfigOption(
+        "high_threshold_branch",
+        ["--high-threshold-branch", "--html-high-threshold-branch"],
+        group="output_options",
+        type=check_percentage,
+        metavar="HIGH_BRANCH",
+        help="If the coverage is below HIGH_BRANCH, the value is marked "
+        "as medium coverage in the report. "
+        "HIGH_BRANCH has to be greater than or equal to value of --medium-threshold-branch. "
+        "If HIGH_BRANCH is equal to value of --medium-threshold-branch the report has "
+        "only high and low coverage. Default is taken from --high-threshold.",
+        default=None,
+    ),
+    GcovrConfigOption(
+        "medium_threshold_line",
+        ["--medium-threshold-line", "--html-medium-threshold-line"],
+        group="output_options",
+        metavar="MEDIUM_LINE",
+        type=check_percentage,
+        help="If the coverage is below MEDIUM_LINE, the value is marked "
+        "as low coverage in the report. "
+        "MEDIUM_LINE has to be lower than or equal to value of --high-threshold-line "
+        "and greater than 0. "
+        "If MEDIUM_LINE is equal to value of --medium-threshold-line the report has "
+        "only high and low coverage. Default is taken from --medium-threshold.",
+        default=None,
+    ),
+    GcovrConfigOption(
+        "high_threshold_line",
+        ["--high-threshold-line", "--html-high-threshold-line"],
+        group="output_options",
+        type=check_percentage,
+        metavar="HIGH_LINE",
+        help="If the coverage is below HIGH_LINE, the value is marked "
+        "as medium coverage in the report. "
+        "HIGH_LINE has to be greater than or equal to value of --medium-threshold-line. "
+        "If HIGH_LINE is equal to value of --medium-threshold-line the report has "
+        "only high and low coverage. Default is taken from --high-threshold.",
+        default=None,
+    ),
     GcovrConfigOption(
         "fail_under_line",
         ["--fail-under-line"],
@@ -525,11 +623,11 @@ GCOVR_CONFIG_OPTIONS = [
         action="store_true",
     ),
     GcovrConfigOption(
-        "exclude_calls",
+        "show_calls",
         ["--calls"],
         group="output_options",
         help="Report the calls coverage. For HTML and the summary report.",
-        action="store_false",
+        action="store_true",
     ),
     GcovrConfigOption(
         "sort_branches",
@@ -553,7 +651,7 @@ GCOVR_CONFIG_OPTIONS = [
             "The secondary sort key (if values are identical) is always the filename (ascending order). "
             "For CSV, HTML, JSON, LCOV and text report."
         ),
-        choices=["filename", "uncovered-number", "uncovered-percent"],
+        choices=("filename", "uncovered-number", "uncovered-percent"),
         default="filename",
     ),
     GcovrConfigOption(
@@ -597,7 +695,22 @@ GCOVR_CONFIG_OPTIONS = [
         default=source_date_epoch() or datetime.datetime.now(),
     ),
     GcovrConfigOption(
-        "filter",
+        "include_search_filter",
+        ["-i", "--include"],
+        group="filter_options",
+        help=(
+            "Include source files that match this filter. "
+            "This is to ensure that files are in report even "
+            "if no coverage data is found. Files are searched "
+            "recursive from the --root directory. "
+            "Can be specified multiple times."
+        ),
+        action="append",
+        type=FilterOption,
+        default=[],
+    ),
+    GcovrConfigOption(
+        "include_filter",
         ["-f", "--filter"],
         group="filter_options",
         help=(
@@ -612,7 +725,7 @@ GCOVR_CONFIG_OPTIONS = [
         default=[],
     ),
     GcovrConfigOption(
-        "exclude",
+        "exclude_filter",
         ["-e", "--exclude"],
         group="filter_options",
         help=(
@@ -620,7 +733,7 @@ GCOVR_CONFIG_OPTIONS = [
             "Can be specified multiple times."
         ),
         action="append",
-        type=FilterOption.NonEmpty,
+        type=NonEmptyFilterOption,
         default=[],
     ),
     GcovrConfigOption(
@@ -628,19 +741,30 @@ GCOVR_CONFIG_OPTIONS = [
         ["--merge-mode-functions"],
         metavar="MERGE_MODE",
         group="gcov_options",
-        choices=[
+        choices=(
             "strict",
             "merge-use-line-0",
             "merge-use-line-min",
             "merge-use-line-max",
             "separate",
-        ],
+        ),
         default="strict",
         help=(
-            "The merge mode for functions coverage from different gcov files for same sourcefile."
+            "The merge mode for functions coverage from different gcov files for same sourcefile. "
             "Default is '{default!s}'."
         ),
     ),
+    GcovrConfigOption(
+        "merge_lines",
+        ["--merge-lines"],
+        group="gcov_options",
+        help=(
+            "Merge line coverage for same line coming from different functions, "
+            "e.g. template instances. The branches, conditions and calls are merged "
+            "accordingly."
+        ),
+        action="store_true",
+    ),
     GcovrConfigOption(
         "exclude_internal_functions",
         ["--include-internal-functions"],
@@ -689,7 +813,7 @@ GCOVR_CONFIG_OPTIONS = [
         action="store_true",
     ),
     GcovrConfigOption(
-        "exclude_functions",
+        "exclude_function",
         ["--exclude-function"],
         help=(
             "Exclude coverage of functions. If function starts and end "
@@ -704,14 +828,18 @@ GCOVR_CONFIG_OPTIONS = [
     GcovrConfigOption(
         "exclude_lines_by_pattern",
         ["--exclude-lines-by-pattern"],
-        help="Exclude lines that match this regex.",
+        help="Exclude lines that match this regex. The regex must match the start of the line.",
+        action="append",
         type=str,
+        default=[],
     ),
     GcovrConfigOption(
         "exclude_branches_by_pattern",
         ["--exclude-branches-by-pattern"],
-        help="Exclude branches that match this regex.",
+        help="Exclude branches that match this regex. The regex must match the start of the line.",
+        action="append",
         type=str,
+        default=[],
     ),
     GcovrConfigOption(
         "exclude_pattern_prefix",
@@ -723,6 +851,12 @@ GCOVR_CONFIG_OPTIONS = [
         type=str,
         default=r"[GL]COVR?",
     ),
+    GcovrConfigOption(
+        "warn_excluded_lines_with_hits",
+        ["--warn-excluded-lines-with-hits"],
+        help="Print a warning if a line excluded by comments has a hit counter != 0.",
+        action="store_true",
+    ),
     GcovrConfigOption(
         "search_paths",
         config="search-path",
@@ -779,15 +913,15 @@ def parse_config_file(
     test.cfg: 7: optional = spaces
     """
 
+    def error(pattern: str, *args: object, **kwargs: object) -> SyntaxError:
+        # pylint: disable=cell-var-from-loop
+        message = pattern.format(*args, **kwargs)
+        message += f"\non this line: {line}"
+        return SyntaxError(": ".join([filename, str(lineno), message]))
+
     for lineno, line in enumerate(open_file, first_lineno):
         line = line.rstrip()
 
-        def error(pattern: str, *args, **kwargs):
-            # pylint: disable=cell-var-from-loop
-            message = pattern.format(*args, **kwargs)
-            message += "\non this line: " + line
-            return SyntaxError(": ".join([filename, str(lineno), message]))
-
         # strip (trailing) comments
         line = CONFIG_HASH_COMMENT.sub("", line)
 
@@ -820,7 +954,7 @@ def parse_config_file(
 
 
 def config_entries_from_dict(
-    config: Dict[str, Any],
+    config: dict[str, Any],
     filename: str,
 ) -> Iterable[ConfigEntry]:
     r"""
@@ -868,7 +1002,7 @@ class ConfigEntry:
     lineno: Optional[int] = None
     """Line of the entry in the config file, for error messages."""
 
-    def __str__(self):
+    def __str__(self) -> str:
         r"""
         Display the config entry.
 
@@ -906,7 +1040,7 @@ class ConfigEntry:
             return False
         raise self.error('boolean option must be "yes" or "no"')
 
-    def error(self, pattern: str, *args, **kwargs) -> ValueError:
+    def error(self, pattern: str, *args: object, **kwargs: object) -> ValueError:
         r"""
         Format but NOT RAISE a ValueError.
 
