@@ -2,8 +2,8 @@
 
 #  ************************** Copyrights and license ***************************
 #
-# This file is part of gcovr 8.3, a parsing and reporting tool for gcov.
-# https://gcovr.com/en/8.3
+# This file is part of gcovr 8.4, a parsing and reporting tool for gcov.
+# https://gcovr.com/en/8.4
 #
 # _____________________________________________________________________________
 #
@@ -27,19 +27,14 @@ import subprocess  # nosec # Commands are trusted.
 from threading import Lock
 from typing import Any, Callable, Optional
 
-from ...coverage import CoverageContainer
+from ...data_model.container import CoverageContainer
+from ...data_model.merging import get_merge_mode_from_options
 from ...decision_analysis import DecisionParser
 from ...exclusions import (
     apply_all_exclusions,
     get_exclusion_options_from_options,
 )
 from ...filter import Filter, is_file_excluded
-from ...merging import (
-    GcovrMergeAssertionError,
-    get_merge_mode_from_options,
-    insert_file_coverage,
-    merge_covdata,
-)
 from ...options import Options
 from ...utils import (
     commonpath,
@@ -63,6 +58,7 @@ output_error_re = re.compile(
     r"(?:[Cc](?:annot|ould not) open output file|Operation not permitted|Permission denied|Read-only file system)"
 )
 unknown_cla_re = re.compile(r"Unknown command line argument")
+version_mismatch_re = re.compile(r":version '[']+', prefer '[']+'")
 
 
 def read_report(options: Options) -> CoverageContainer:
@@ -83,7 +79,7 @@ def read_report(options: Options) -> CoverageContainer:
             options.search_paths.append(options.gcov_objdir)
 
     for search_path in options.search_paths:
-        datafiles.update(find_files(search_path, options.gcov_exclude_dirs))
+        datafiles.update(find_files(search_path, options.gcov_exclude_directory))
 
     # Get coverage data
     with Workers(
@@ -91,16 +87,19 @@ def read_report(options: Options) -> CoverageContainer:
         lambda: {"covdata": CoverageContainer(), "to_erase": set(), "options": options},
     ) as pool:
         LOGGER.debug(f"Pool started with {pool.size()} threads")
-        for file_ in sorted(datafiles):
-            pool.add(process_file, file_)
-        contexts = pool.wait()
+        for filename in sorted(datafiles):
+            pool.add(process_file, filename)
+        try:
+            contexts = pool.wait()
+        except KeyboardInterrupt as exc:
+            # Stop the pool if Ctrl+C is pressed
+            pool.drain()
+            raise exc from None
 
     to_erase = set()
     covdata = CoverageContainer()
     for context in contexts:
-        covdata = merge_covdata(
-            covdata, context["covdata"], get_merge_mode_from_options(options)
-        )
+        covdata.merge(context["covdata"], get_merge_mode_from_options(options))
         to_erase.update(context["to_erase"])
 
     for filepath in to_erase:
@@ -111,11 +110,11 @@ def read_report(options: Options) -> CoverageContainer:
 
 
 def find_existing_gcov_files(
-    search_path: str, exclude_dirs: list[re.Pattern[str]]
+    search_path: str, gcov_exclude_directory: list[re.Pattern[str]]
 ) -> list[str]:
     """Find .gcov and .gcov.json.gz files under the given search path."""
     if os.path.isfile(search_path):
-        LOGGER.debug(f"Using given file {search_path}")
+        LOGGER.debug(f"Using given gcov file {search_path}")
         gcov_files = [search_path]
     else:
         LOGGER.debug(f"Scanning directory {search_path} for gcov files...")
@@ -124,14 +123,16 @@ def find_existing_gcov_files(
                 lambda fname: re.compile(r".*\.gcov(?:\.json\.gz)?$").match(fname)
                 is not None,
                 search_path,
-                exclude_dirs=exclude_dirs,
+                gcov_exclude_directory=gcov_exclude_directory,
             )
         )
         LOGGER.debug(f"Found {len(gcov_files)} files (and will process all of them)")
     return gcov_files
 
 
-def find_datafiles(search_path: str, exclude_dirs: list[re.Pattern[str]]) -> list[str]:
+def find_datafiles(
+    search_path: str, gcov_exclude_directory: list[re.Pattern[str]]
+) -> list[str]:
     """Find .gcda and .gcno files under the given search path.
 
     The .gcno files will *only* produce uncovered results.
@@ -140,7 +141,9 @@ def find_datafiles(search_path: str, exclude_dirs: list[re.Pattern[str]]) -> lis
     So we ONLY return them if there's no corresponding .gcda file.
     """
     if os.path.isfile(search_path):
-        LOGGER.debug(f"Using given file {search_path}")
+        LOGGER.debug(
+            f"Using given {os.path.splitext(search_path)[1][1:]} file {search_path}"
+        )
         files = [search_path]
     else:
         LOGGER.debug(f"Scanning directory {search_path} for gcda/gcno files...")
@@ -148,7 +151,7 @@ def find_datafiles(search_path: str, exclude_dirs: list[re.Pattern[str]]) -> lis
             search_file(
                 lambda fname: re.compile(r".*\.gc(da|no)$").match(fname) is not None,
                 search_path,
-                exclude_dirs=exclude_dirs,
+                gcov_exclude_directory=gcov_exclude_directory,
             )
         )
     gcda_files = []
@@ -185,29 +188,30 @@ def process_gcov_json_data(
         gcov_json_data = json_loads(fh_in.read())
 
     coverage = json.parse_coverage(
+        data_fname,
         gcov_json_data=gcov_json_data,
-        include_filters=options.filter,
-        exclude_filters=options.exclude,
+        include_filter=options.include_filter,
+        exclude_filter=options.exclude_filter,
         ignore_parse_errors=options.gcov_ignore_parse_errors,
         suspicious_hits_threshold=options.gcov_suspicious_hits_threshold,
         source_encoding=options.source_encoding,
-        data_fname=data_fname,
     )
 
-    for file_cov, source_lines in coverage:
-        LOGGER.debug(f"Apply exclusions for {file_cov.filename}")
+    merge_options = get_merge_mode_from_options(options)
+    for filecov, source_lines in coverage:
+        LOGGER.debug(f"Apply exclusions for {filecov.filename}")
         apply_all_exclusions(
-            file_cov,
+            filecov,
             lines=source_lines,
             options=get_exclusion_options_from_options(options),
         )
 
         if options.show_decision:
-            decision_parser = DecisionParser(file_cov, source_lines)
+            decision_parser = DecisionParser(filecov, source_lines)
             decision_parser.parse_all_lines()
 
-        LOGGER.debug(f"Merge coverage data for {file_cov.filename}")
-        insert_file_coverage(covdata, file_cov, get_merge_mode_from_options(options))
+        LOGGER.debug(f"Merge coverage data for {filecov.filename}")
+        covdata.insert_file_coverage(filecov, merge_options)
 
 
 #
@@ -228,7 +232,9 @@ def process_gcov_text_data(
 
     # Find the source file
     metadata = text.parse_metadata(
-        lines, suspicious_hits_threshold=options.gcov_suspicious_hits_threshold
+        data_fname,
+        lines,
+        suspicious_hits_threshold=options.gcov_suspicious_hits_threshold,
     )
     source = metadata.get("Source")
     if source is None:
@@ -251,29 +257,30 @@ def process_gcov_text_data(
         current_dir=current_dir,
     )
 
-    if is_file_excluded(fname, options.filter, options.exclude):
+    if is_file_excluded(fname, options.include_filter, options.exclude_filter):
         return
 
     LOGGER.debug(f"Parsing coverage data for file {fname}")
     key = os.path.normpath(fname)
 
-    coverage, source_lines = text.parse_coverage(
+    filecov, source_lines = text.parse_coverage(
+        set([(gcda_fname, data_fname) if gcda_fname else (data_fname,)]),
         lines,
         filename=key,
-        data_filename=gcda_fname or data_fname,
         ignore_parse_errors=options.gcov_ignore_parse_errors,
         suspicious_hits_threshold=options.gcov_suspicious_hits_threshold,
     )
 
     LOGGER.debug(f"Apply exclusions for {fname}")
-    apply_all_exclusions(coverage, lines=source_lines, options=options)  # type: ignore [arg-type]
+    apply_all_exclusions(filecov, lines=source_lines, options=options)  # type: ignore [arg-type]
 
     if options.show_decision:
-        decision_parser = DecisionParser(coverage, source_lines)
+        decision_parser = DecisionParser(filecov, source_lines)
         decision_parser.parse_all_lines()
 
-    LOGGER.debug(f"Merge coverage data for {fname}")
-    insert_file_coverage(covdata, coverage, get_merge_mode_from_options(options))
+    merge_mode = get_merge_mode_from_options(options)
+    LOGGER.debug(f"Merge coverage data for {fname} using {str(merge_mode)}")
+    covdata.insert_file_coverage(filecov, merge_mode)
 
 
 def guess_source_file_name(
@@ -466,31 +473,27 @@ def process_datafile(
             potential_wd.append(wd)
             wd = os.path.dirname(wd)
 
-    try:
-        for wd in potential_wd:
-            done = run_gcov_and_process_files(
-                abs_filename,
-                covdata,
-                options=options,
-                error=errors.append,
-                chdir=wd,
-            )
+    for wd in potential_wd:
+        done = run_gcov_and_process_files(
+            abs_filename,
+            covdata,
+            options=options,
+            error=errors.append,
+            chdir=wd,
+        )
 
-            if options.gcov_delete:
-                if not abs_filename.endswith("gcno"):
-                    to_erase.add(abs_filename)
+        if options.gcov_delete:
+            if not abs_filename.endswith("gcno"):
+                to_erase.add(abs_filename)
 
-            if done:
-                return
-    # This exception fails fast
-    except GcovrMergeAssertionError as exc:
-        errors += str(exc).split("\n")
+        if done:
+            return
 
     errors_output = "\n\t".join(errors)
     errors_output = (
         f"GCOV produced the following errors processing {filename}:\n"
         f"\t{errors_output}\n"
-        "\t(gcovr could not infer a working directory that resolved it.)\n"
+        "\t(GCOVR could not infer a working directory that resolved it.)\n"
         "To ignore this error use option --gcov-ignore-errors=no_working_dir_found."
     )
     LOGGER.error(errors_output)
@@ -539,7 +542,6 @@ class GcovProgram:
     __cmd_split = list[str]()
     __default_options = list[str]()
     __exitcode_to_ignore = list[int]([0])
-    __use_json_format_if_available: bool = True
     __help_output: str = ""
     __version_output: str = ""
 
@@ -555,9 +557,8 @@ class GcovProgram:
         def __exit__(self, *_: Any) -> None:
             self.lock.release()
 
-    def __init__(self, cmd: str, options: Options) -> None:
+    def __init__(self, cmd: str) -> None:
         with GcovProgram.LockContext(GcovProgram.__lock):
-            GcovProgram.__use_json_format_if_available = options.exclude_calls
             if not GcovProgram.__cmd:
                 GcovProgram.__cmd = cmd
                 # If the first element of cmd - the executable name - has embedded spaces
@@ -578,20 +579,17 @@ class GcovProgram:
                     "--all-blocks",
                 ]
 
-                if (
-                    GcovProgram.__use_json_format_if_available
-                    and self.__check_gcov_help_content("--json-format")
-                ):
+                if self.__check_gcov_help_content("--json-format"):
                     if self.__check_gcov_version_content(
                         f"JSON format version: {json.GCOV_JSON_VERSION}"
                     ):
                         LOGGER.debug("GCOV capabilities: JSON format available.")
                         GcovProgram.__default_options.append("--json-format")
-                        if self.__check_gcov_help_content("--condition"):
+                        if self.__check_gcov_help_content("--conditions"):
                             LOGGER.debug(
                                 "GCOV capabilities: Condition coverage available."
                             )
-                            GcovProgram.__default_options.append("--condition")
+                            GcovProgram.__default_options.append("--conditions")
                     else:
                         LOGGER.debug(
                             "GCOV capabilities: Unsupported JSON format detected."
@@ -728,7 +726,7 @@ class GcovProgram:
         gcov_process = self.__get_gcov_process(args, **kwargs)
         out, err = gcov_process.communicate()
         LOGGER.debug(
-            f"GCOV return code was {gcov_process.returncode}, stderr was:\n{err}<<"
+            f"GCOV return code was {gcov_process.returncode}, stdout was:\n{out}<<, stderr was:\n{err}<<"
         )
         if gcov_process.returncode < 0:
             raise RuntimeError(
@@ -755,17 +753,28 @@ def run_gcov_and_process_files(
     chdir: str,
 ) -> bool:
     """Run GCOV tool and process the output files."""
-    filename = None
-    out = None
-    err = None
-    try:
-        gcov_cmd = GcovProgram(options.gcov_cmd, options)
-        gcov_cmd.identify_and_cache_capabilities()
 
-        # ATTENTION:
-        # This lock is essential for parallel processing because without
-        # this there can be name collisions for the generated output files.
-        with locked_directory(chdir):
+    done = False
+
+    # ATTENTION:
+    # This lock is essential for parallel processing because without
+    # this there can be name collisions for the generated output files.
+    with locked_directory(chdir):
+
+        def remove_existing_files(files: list[str]) -> None:
+            """Remove the existing files from the given list."""
+            for filepath in sorted(files):
+                if os.path.exists(filepath):
+                    os.remove(filepath)
+
+        filename = None
+        out = None
+        err = None
+        active_gcov_files = set[str]()
+        try:
+            gcov_cmd = GcovProgram(options.gcov_cmd)
+            gcov_cmd.identify_and_cache_capabilities()
+
             filename = abs_filename
             # Use try catch because the relpath can fail on Windows for different drives.
             # Do not know how to force this exception therefore ignore coverage.
@@ -791,14 +800,19 @@ def run_gcov_and_process_files(
             # find the files that gcov created
             active_gcov_files, all_gcov_files = select_gcov_files_from_stdout(
                 out,
-                gcov_filter=options.gcov_filter,
-                gcov_exclude=options.gcov_exclude,
+                include_filter=options.gcov_include_filter,
+                exclude_filter=options.gcov_exclude_filter,
                 chdir=chdir,
             )
+            # Remove the not used files
+            remove_existing_files(list(all_gcov_files - active_gcov_files))
 
             if unknown_cla_re.search(err):
                 # gcov tossed errors: throw exception
                 raise RuntimeError(f"Error in gcov command line: {err}")
+            if version_mismatch_re.search(err):
+                # gcov tossed errors: throw exception
+                raise RuntimeError(f"Version mismatch in gcc/gcov: {err}")
 
             ignore_source_errors = options.gcov_ignore_errors is not None and any(
                 v in options.gcov_ignore_errors for v in ["all", "source_not_found"]
@@ -838,41 +852,42 @@ def run_gcov_and_process_files(
                         raise RuntimeError(
                             f"Unknown gcov output format {gcov_filename}."
                         )
+
                 done = True
 
+        except RuntimeError as exc:
+            # If we got an merge assertion error we must end the processing
+            done = False
+            error(
+                f"Trouble processing {abs_filename!r} with working directory {chdir!r}.\n"
+                f"Stdout of gcov was >>{out}<< End of stdout\n"
+                f"Stderr of gcov was >>{err}<< End of stderr\n"
+                f"Exception was >>{str(exc)}<< End of stderr\n"
+                f"Current processed gcov file was {filename!r}.\n"
+                "Use option --verbose to get extended information."
+            )
+        finally:
             if options.gcov_keep and done:
+                # Keep the files with unique names
                 basename = os.path.basename(abs_filename)
-                for file in active_gcov_files:
-                    directory, filename = os.path.split(file)
-                    os.replace(file, os.path.join(directory, f"{basename}.{filename}"))
-
-            for filepath in (
-                list(all_gcov_files - active_gcov_files)
-                if options.gcov_keep and done
-                else all_gcov_files
-            ):
-                if os.path.exists(filepath):
-                    os.remove(filepath)
-
-    except RuntimeError as exc:
-        # If we got an merge assertion error we must end the processing
-        done = False
-        error(
-            f"Trouble processing {abs_filename!r} with working directory {chdir!r}.\n"
-            f"Stdout of gcov was >>{out}<< End of stdout\n"
-            f"Stderr of gcov was >>{err}<< End of stderr\n"
-            f"Exception was >>{str(exc)}<< End of stderr\n"
-            f"Current processed gcov file was {filename!r}.\n"
-            "Use option --verbose to get extended information."
-        )
+                for gcov_filename in active_gcov_files:
+                    if os.path.exists(gcov_filename):
+                        directory, filename = os.path.split(gcov_filename)
+                        os.replace(
+                            gcov_filename,
+                            os.path.join(directory, f"{basename}.{filename}"),
+                        )
+            else:
+                # Remove the used files
+                remove_existing_files(list(active_gcov_files))
 
     return done
 
 
 def select_gcov_files_from_stdout(
     out: str,
-    gcov_filter: list[Filter],
-    gcov_exclude: list[Filter],
+    include_filter: list[Filter],
+    exclude_filter: list[Filter],
     chdir: str,
 ) -> tuple[set[str], set[str]]:
     """Parse the output to get the list of files to use and all files (unfiltered)."""
@@ -888,7 +903,7 @@ def select_gcov_files_from_stdout(
         full = os.path.join(chdir, fname)
         all_files.add(full)
 
-        if is_file_excluded(fname, gcov_filter, gcov_exclude):
+        if is_file_excluded(fname, include_filter, exclude_filter):
             continue
 
         active_files.add(full)
@@ -903,7 +918,9 @@ def process_existing_gcov_file(
     filename: str, covdata: CoverageContainer, options: Options, to_erase: set[str]
 ) -> None:
     """Process an existing GCOV filename."""
-    if is_file_excluded(filename, options.gcov_filter, options.gcov_exclude):
+    if is_file_excluded(
+        filename, options.gcov_include_filter, options.gcov_exclude_filter
+    ):
         LOGGER.debug(f"Excluding gcov file: {filename}")
 
     if filename.endswith(".gcov"):
